/*
	bhtxt.c

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "borhelp.h"

typedef struct {
	char		*mem;
	unsigned	length;
	int			flag;
} nibble;


int
getstamp(char *dest, int limit, FILE *fi)
{
	int		result = 0;
	int		c;
	
	while(-1 != (c = fgetc(fi))) {
		if (ferror(fi)) { result = 0; break; }
		if (c == '\0') break;
		if (result < limit) {
			dest[result] = c;
		}
		result++;
	}
	
	dest[result] = '\0';
	if (result) fgetc(fi);	/* drop EOF */
	return result;
}

int
getasciz(char *dest, int limit, FILE *fi)
{
	int		result = 0;
	int		c;
	
	while(-1 != (c = fgetc(fi))) {
		if (ferror(fi)) { result = 0; break; }
		if (c == '\0') break;
		if (result < limit) {
			dest[result] = c;
		}
		result++;
	}
	
	dest[result] = '\0';
	return result;
}

void
dumpcomprec(FILE *fo, bh_comprec *comp)
{
	int		i;
	
	fprintf(fo, ";COMMENT compression method %02x", comp->method);
	switch(comp->method) {
		case METHOD_NONE:
			fputs(" (NON COMPRESSION)", fo);
			break;
		case METHOD_RLE:
			fputs(" (RUN-LENGTH)", fo);
			break;
		case METHOD_NIBBLE:
			fputs(" (NIBBLE)", fo);
			break;
		default:
			fputs(" (unknown)", fo);
	}
	
	fprintf(fo, "\n;COMMENT ");
	
	for(i=0; i<14; i++) {
		fprintf(fo, " %02x", comp->table[i]);
	}
	fputc('\n', fo);
	
	return ;
}


unsigned char
getnibble(nibble *n)
{
	unsigned char	c;
	
	c = *(n->mem);
	if (n->flag) {
		c >>= 4;
		n->flag = 0;
		(n->length)--;
		(n->mem)++;
	} else {
		c &= 15;
		n->flag = 1;
	}
	
	return c;
}


unsigned char
getnextchar(nibble * n, bh_comprec * comp, int *repeat)
{
	unsigned char	c, cl, ch;
	
	*repeat = 1;
	c = getnibble(n);
	
	switch(c) {
		case NIBBLE_LITERAL:
			cl = getnibble(n);
			ch = getnibble(n);
			c = (ch << 4) + cl;
			break;
		
		case NIBBLE_REPEAT:
			cl = getnibble(n) + 2;
			c = getnextchar(n, comp, repeat);
			*repeat = cl;
			break;
		
		default:
			c = comp->table[c];
			break;
	}
	
	return c;
}


void
dumptext(FILE *ft, char *m, unsigned length, bh_comprec * comp, long *contents)
{
	nibble		nib;
	unsigned char	c;
	int		rep;
	
	nib.flag = 0;
	nib.mem = m;
	nib.length = length;
	
	fprintf(ft, "\n;COMMENT ----------------------------------------\n");
	fprintf(ft, ";SCREEN hl_%ld\n\n", *contents);
	
	while(nib.length) {
		c = getnextchar(&nib, comp, &rep);
		if (c == 0) c = '\n';
		while(rep--) {
			fputc(c, ft);
		}
	}
	
	fprintf(ft, "\n\n;ENDSCREEN\n\n");
	(*contents)++;
}


void
dumpfilehdr(FILE *fo, bh_filehdr * f)
{
	fprintf(fo, ";COMMENT options 0x%04x\n", f->options);
	if (f->options & 4) {
		fprintf(fo, ";CASESENSE\n");
	}
	fprintf(fo, ";MAININDEX %d\n", f->main_idx);
	fprintf(fo, ";COMMENT maxscreensize %d\n", f->max_screen_size);
	fprintf(fo, ";WIDTH %d\n", f->width);
	fprintf(fo, ";HEIGHT %d\n", f->height);
	fprintf(fo, ";LMARGIN %d\n", f->margin_left);
	
}


void
dumpbyte(FILE *fo, char *m, int length)
{
	int		i;
	int		cnt = 0;
	
	while(length > cnt) {
		fprintf(fo, ";COMMENT  %04X", cnt);
		for(i=0; i<16; i++) {
			if (length <= cnt) break;
			fprintf(fo, " %02x", *m++);
			cnt++;
		}
		fputc('\n', fo);
	}
	
}


int
dumprec(FILE *fo, FILE *fi, long * contents)
{
	bh_rechdr	hdr;
	bh_comprec	comp;
	char *		rec;
	
	hdr.length = 0;
	if (fread(&hdr, sizeof(hdr), 1, fi) == 0) return 1;
	
	fprintf(fo, "\n;COMMENT record %d, length %d\n", hdr.rec_type, hdr.length);
	
	if (hdr.length == 0) return 0;
	
	rec = (char *)malloc(hdr.length);
	if (!rec) { fprintf(stderr, "cant allocate memories.\n"); return 1; }
	if (fread(rec, hdr.length, 1, fi) == 0) {
		fprintf(stderr, "cant read a record\n");
		free(rec);
		return 1;
	}
	
	switch(hdr.rec_type) {
		case RT_FILEHEADER:
			dumpfilehdr(fo, (bh_filehdr *)rec);
			break;
		case RT_CONTEXT:
			break;
		case RT_TEXT:
			dumptext(fo, rec, hdr.length, &comp, contents);
			break;
		case RT_KEYWORD:
			break;
		case RT_INDEX:
			break;
		case RT_COMPRESSION:
			memcpy(&comp, rec, sizeof(bh_comprec));
			dumpcomprec(fo, &comp);
			break;
		
		default:
			break;
	}
	
	free(rec);
	return 0;
}


void
dumpversion(FILE *fo, FILE * fi)
{
	bh_version	ver;
	
	fread(&ver, sizeof(ver), 1, fi);
	fprintf(fo, ";VERSION %d\n", ver.text_version);
	fprintf(fo, ";COMMENT FORMATVERSION %d\n", ver.format_version);
}


int
dumpstamp(FILE *fo, FILE *fi)
{
	char	stamp[128];
	char	sign[16];
	
	getstamp(stamp, sizeof(stamp), fi);
	fprintf(fo, ";STAMP %s\n", stamp);
	getasciz(sign, sizeof(sign), fi);
	fprintf(fo, ";SIGNATURE %s\n", sign);
	
	return (strcmp(BH_SIGNATURE, sign) == 0);
}


int
dumphelp(FILE *fo, FILE *fi)
{
	long	contents = 1;
	
	if (!dumpstamp(fo, fi)) {
		fprintf(stderr, "Invalid signature");
		return 1;
	}
	dumpversion(fo, fi);
	
	while(!feof(fi)) {
		dumprec(fo, fi, &contents);
	}
	
	return 0;
}


int
main(int argc, char ** argv)
{
	FILE		*fi, *fo;
	
	if (argc <= 1) {
		puts("bhtxt helpfile [txtfile]");
		return 1;
	}
	
	fi = fopen(argv[1], "rb");
	if (!fi) { printf("Can't open %s\n", argv[1]); return 1; }
	
	if (argc <= 2) fo = stdout;
	else {
		fo = fopen(argv[2], "wt");
	}
	
	dumphelp(fo, fi);
	
	if (fo != stdout) fclose(fo);
	fclose(fi);
	return 0;
}

